from surpyvor import utils, parse_arguments
import sys


def main():
    args = parse_arguments.get_args()
    utils.test_dependencies()
    if args.command == "merge":
        sv_merge(
            samples=args.variants,
            distance=args.distance,
            callers=args.callers,
            require_type=not args.ignore_type,
            require_strand=args.strand,
            estimate_distance=args.estimate_distance,
            minlength=args.minlength,
            output=args.output,
            verbose=args.verbose,
        )
    elif args.command == "highsens":
        sv_merge(
            samples=[utils.vcf_concat(args.variants)],
            distance=args.distance,
            callers=1,
            require_type=True,
            require_strand=False,
            estimate_distance=False,
            minlength=args.minlength,
            output=args.output,
            verbose=args.verbose,
        )
    elif args.command == "highconf":
        sv_merge(
            samples=args.variants,
            distance=args.distance,
            callers=len(args.variants),
            require_type=True,
            require_strand=args.strand,
            estimate_distance=False,
            minlength=args.minlength,
            output=args.output,
            verbose=args.verbose,
        )
    elif args.command == "prf":
        precision_recall_fmeasure(args)
    elif args.command == "upset":
        upset(args)
    elif args.command == "venn":
        venn(args)
    elif args.command == "haplomerge":
        haplomerge(args)
    elif args.command == "lengthplot":
        lengthplot(args)
    elif args.command == "minlen":
        minlen(args)
    elif args.command == "svlentruncate":
        svlentruncate(args)
    elif args.command == "fixvcf":
        utils.fix_vcf(args.vcf, args.output, args.fai, jasmine=args.jasmine)
    elif args.command == "purge2d":
        purge2d(args)
    elif args.command == "carrierplot":
        from surpyvor.plots import carrierplot

        carrierplot(args)
    elif args.command == "varcount":
        from surpyvor.plots import num_variants_per_sample

        num_variants_per_sample(args.variants, args.plotout, args.countsout)
    elif args.command == "fixref":
        from surpyvor.fixref import fixref

        fixref(args.variants, args.fasta)


def sv_merge(
    samples,
    distance,
    callers,
    require_type,
    require_strand,
    estimate_distance,
    minlength,
    output,
    verbose=False,
):
    """
    Executes SURVIVOR merge, with parameters:
    -samples.fofn (samples, list)
    -distance between calls (distance, int)
    -number of callers to support call (callers, int)
    -require variants to have sampe type (type, boolean)
    -require variants to be on same strand (strand, boolean)
    -estimate distance between calls (estimate_distance, boolean)
    -specify minimal size of SV event (minlength, int)
    """
    import subprocess
    import shlex
    import os
    import tempfile

    fhf, fofn_f = tempfile.mkstemp()
    fhs, interm_out = tempfile.mkstemp(suffix=".vcf")
    with open(fofn_f, "w") as fofn:
        for s in [utils.decompress(s) for s in samples]:
            fofn.write(s + "\n")
    survivor_cmd = "SURVIVOR merge {fof} {dist} {call} {typ} {str} {estm} {ml} {out}".format(
        fof=fofn_f,
        dist=distance,
        call=callers,
        typ=1 if require_type else -1,
        str=1 if require_strand else -1,
        estm=1 if estimate_distance else -1,
        ml=minlength,
        out=interm_out,
    )
    if verbose:
        print("\n\nExecuting:", file=sys.stderr)
        print(survivor_cmd, file=sys.stderr)
    print("Executing SURVIVOR...", end="", flush=True, file=sys.stderr)
    subprocess.call(shlex.split(survivor_cmd), stdout=subprocess.DEVNULL)
    print("DONE", file=sys.stderr)
    if verbose:
        print("\n\nSorting merged vcf file:", file=sys.stderr)
    utils.vcf_sort(interm_out, output)
    os.close(fhf)
    os.close(fhs)


def snv_merge(samples, output, verbose=False):
    import subprocess
    import shlex

    inputfiles = " ".join(samples)
    bcftools_cmd = f"bcftools merge {inputfiles} -o {output}"
    if verbose:
        print("\n\nExecuting:", file=sys.stderr)
        print(bcftools_cmd, file=sys.stderr)
    print("Executing bcftools...", end="", flush=True, file=sys.stderr)
    subprocess.call(shlex.split(bcftools_cmd), stdout=subprocess.DEVNULL)
    print("DONE", file=sys.stderr)


def default_merge(args, variants):
    import tempfile

    if args.keepmerged:
        vcf_out = args.keepmerged
    else:
        _, vcf_out = tempfile.mkstemp()
    if args.snv:
        snv_merge(samples=variants, output=vcf_out, verbose=args.verbose)
    else:
        sv_merge(
            samples=[utils.normalize_vcf(s) for s in variants],
            distance=args.distance,
            callers=1,
            require_type=not args.ignore_type,
            require_strand=False,
            estimate_distance=False,
            minlength=args.minlength,
            output=vcf_out,
            verbose=args.verbose,
        )
    return vcf_out


def precision_recall_fmeasure(args):
    vcf_out = default_merge(args, variants=[args.truth, args.test])
    truth_set, test_set = utils.get_variant_identifiers(
        vcf=vcf_out, ignore_chroms=args.ignore_chroms
    )

    tp = len(truth_set & test_set)
    precision = tp / len(test_set)
    print(f"Precision: {round(precision, ndigits=4)}")
    recall = tp / len(truth_set)
    print(f"Recall: {round(recall, ndigits=4)}")
    fmeasure = 2 * (precision * recall) / (precision + recall)
    print(f"F-measure: {round(fmeasure, ndigits=4)}")

    if args.venn:
        from surpyvor.plots import venn_diagram

        venn_diagram((truth_set, test_set), labels=("Truth", "Test"))
    if args.bar:
        from surpyvor.plots import bar_chart

        bar_chart(vcf_out)
    if args.matrix:
        utils.confusion_matrix(vcf_out, names=["truth", "test"])


def upset(args):
    from surpyvor.plots import upset_plot

    vcf_out = default_merge(args, args.variants)
    upsets = utils.make_sets(vcf=vcf_out, names=args.names or args.variants)
    upset_plot(upsets, outname=args.plotout)


def venn(args):
    from surpyvor.plots import venn_diagram

    vcf_out = default_merge(args, args.variants)
    sets = utils.get_variant_identifiers(
        vcf=vcf_out, ignore_chroms=[], num_samples=len(args.variants)
    )
    venn_diagram(
        sets,
        labels=args.names or args.variants,
        num_samples=len(args.variants),
        outname=args.plotout,
    )


def haplomerge(args):
    from surpyvor import haplomerge as hm

    args.keepmerged = False
    merged = default_merge(args, args.variants)
    hm.merge_split_called_haplotypes(merged, output=args.output, name=args.name)


def purge2d(args):
    from surpyvor import purge2d as p2d

    p2d.process(args.bam, output=args.output)


def lengthplot(args):
    from surpyvor.plots import length_plot

    len_dict = utils.get_svlengths(args.vcf, all=args.all)
    with open(args.counts, "w") as counts:
        counts.write("Number of nucleotides affected by SV:\n")
        for svtype, lengths in len_dict.items():
            counts.write("{}:\t{} variants\t{}bp\n".format(svtype, len(lengths), sum(lengths)))
    length_plot(dict_of_lengths=len_dict, output=args.plotout)


def minlen(args):
    utils.filter_vcf(args.vcf, output=args.output, minlength=args.length, suffix="filtered")


def svlentruncate(args):
    utils.filter_vcf(args.vcf, output=args.output, truncate_svlen=args.length, suffix="truncated")


if __name__ == "__main__":
    main()



from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from .version import __version__
import sys
from os import path


def get_args():
    parser = ArgumentParser(
        description="A wrapper around SURVIVOR, with convenience functions.",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="surpyvor: {}, SURVIVOR {}".format(__version__, get_survivor_version()),
        help="Print version and quit.",
    )

    parser.add_argument(
        "--snv",
        help="vcf file has SNVs instead of SVs",
        action="store_true",
        default=False,
    )

    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "--verbose", help="Print out more information while running.", action="store_true"
    )
    subparsers = parser.add_subparsers(dest="command", title="[sub-commands]")
    merge = subparsers.add_parser("merge", help="merging vcf files of SVs", parents=[parent_parser])
    merge_req = merge.add_argument_group("required arguments")
    merge_req.add_argument("--variants", nargs="+", required=True, help="vcf files to merge")
    merge_opt = merge.add_argument_group("optional arguments")
    merge_opt.add_argument("-o", "--output", help="output file", default="stdout")
    merge_opt.add_argument(
        "-d", "--distance", type=int, default=500, help="distance between variants to merge"
    )
    merge_opt.add_argument(
        "-l", "--minlength", type=int, default=50, help="Minimum length of variants to consider"
    )
    merge_opt.add_argument(
        "-c",
        "--callers",
        type=int,
        default=1,
        help="Minimum number of callers to support a variant",
    )
    merge_opt.add_argument(
        "-i",
        "--ignore_type",
        help="Ignore the type of the structural variant",
        action="store_true",
        default=False,
    )
    merge_opt.add_argument(
        "-s", "--strand", action="store_true", default=False, help="Take strand into account"
    )
    merge_opt.add_argument(
        "-e",
        "--estimate_distance",
        action="store_true",
        default=False,
        help="Estimate distance between calls",
    )

    highsens = subparsers.add_parser(
        "highsens", help="get union of SV vcfs", parents=[parent_parser]
    )
    highsens_req = highsens.add_argument_group("required arguments")
    highsens_req.add_argument("--variants", nargs="+", required=True, help="vcf files to merge")
    highsens_opt = highsens.add_argument_group("optional arguments")
    highsens_opt.add_argument("-o", "--output", help="output file", default="stdout")
    highsens_opt.add_argument(
        "-d", "--distance", type=int, default=100, help="distance between variants to merge"
    )
    highsens_opt.add_argument(
        "-l", "--minlength", type=int, default=50, help="Minimum length of variants to consider"
    )
    highconf = subparsers.add_parser(
        "highconf", help="get intersection of SV vcfs", parents=[parent_parser]
    )
    highconf_req = highconf.add_argument_group("required arguments")
    highconf_req.add_argument("--variants", nargs="+", required=True, help="vcf files to merge")
    highconf_opt = highconf.add_argument_group("optional arguments")
    highconf_opt.add_argument("-o", "--output", help="output file", default="stdout")
    highconf_opt.add_argument(
        "-d", "--distance", type=int, default=500, help="distance between variants to merge"
    )
    highconf_opt.add_argument(
        "-l", "--minlength", type=int, default=50, help="Minimum length of variants to consider"
    )
    highconf_opt.add_argument(
        "-s", "--strand", action="store_true", default=False, help="Take strand into account"
    )
    prf = subparsers.add_parser(
        "prf", help="calculate precision, recall and F-measure", parents=[parent_parser]
    )
    prf_req = prf.add_argument_group("required arguments")
    prf_req.add_argument("--truth", help="vcf containing truth set", required=True)
    prf_req.add_argument("--test", help="vcf containing test set", required=True)
    prf_opt = prf.add_argument_group("optional arguments")
    prf_opt.add_argument(
        "-d", "--distance", help="maximum distance between test and truth call", default=500
    )
    prf_opt.add_argument(
        "--minlength", help="Minimum length of SVs to be taken into account", default=50
    )
    prf_opt.add_argument(
        "-i",
        "--ignore_type",
        help="Ignore the type of the structural variant",
        action="store_true",
        default=False,
    )
    prf_opt.add_argument(
        "--ignore_chroms",
        help="Chromosomes to ignore for prf calculation.",
        nargs="*",
        default=["chrEBV"],
    )
    prf_opt.add_argument("--keepmerged", help="Save merged vcf file.", default=False)
    prf_opt.add_argument(
        "--bar",
        help="Make stacked bar chart of SV lengths coloured by validation status",
        action="store_true",
    )
    prf_opt.add_argument("--matrix", help="Make a confusion matrix.", action="store_true")
    prf_opt.add_argument("--venn", help="Make a venn diagram.", action="store_true")

    venn = subparsers.add_parser(
        "venn", help="Make venn diagram for 2 or 3 SV vcf files", parents=[parent_parser]
    )
    venn_req = venn.add_argument_group("required arguments")
    venn_req.add_argument(
        "--variants", help="vcfs containing structural variants", required=True, nargs="*"
    )
    venn_opt = venn.add_argument_group("optional arguments")
    venn_opt.add_argument("--names", help="Names of datasets in --variants", nargs="*")
    venn_opt.add_argument(
        "-d", "--distance", help="maximum distance between test and truth call", default=500
    )
    venn_opt.add_argument(
        "--minlength", help="Minimum length of SVs to be taken into account", default=50
    )
    venn_opt.add_argument(
        "-i",
        "--ignore_type",
        help="Ignore the type of the structural variant",
        action="store_true",
        default=False,
    )
    venn_opt.add_argument("--keepmerged", help="Save merged vcf file")
    venn_opt.add_argument("--plotout", help="Name of output plot", default="venn.png")

    upset = subparsers.add_parser(
        "upset", help="Make upset plot for multiple SV vcf files", parents=[parent_parser]
    )
    upset_req = upset.add_argument_group("required arguments")
    upset_req.add_argument(
        "--variants", help="vcfs containing structural variants", required=True, nargs="*"
    )
    upset_opt = upset.add_argument_group("optional arguments")
    upset_opt.add_argument("--names", help="Names of datasets in --variants", nargs="*")
    upset_opt.add_argument(
        "-d", "--distance", help="maximum distance between test and truth call", default=500
    )
    upset_opt.add_argument(
        "--minlength", help="Minimum length of SVs to be taken into account", default=50
    )
    upset_opt.add_argument(
        "-i",
        "--ignore_type",
        help="Ignore the type of the structural variant",
        action="store_true",
        default=False,
    )
    upset_opt.add_argument("--keepmerged", help="Save merged vcf file")
    upset_opt.add_argument("--plotout", help="Name of output plot", default="UpSetPlot.png")

    haplomerge = subparsers.add_parser(
        "haplomerge",
        help="merging vcf files of SVs from two haplotypes",
        formatter_class=ArgumentDefaultsHelpFormatter,
        parents=[parent_parser],
    )
    haplomerge_req = haplomerge.add_argument_group("required arguments")
    haplomerge_req.add_argument("--variants", required=True, nargs="*", help="vcf files to merge")
    haplomerge_opt = haplomerge.add_argument_group("optional arguments")
    haplomerge_opt.add_argument("-o", "--output", help="output file", default="stdout")
    haplomerge_opt.add_argument(
        "-n", "--name", help="name of sample in output VCF", default="stdout"
    )
    haplomerge_opt.add_argument(
        "-d", "--distance", type=int, default=200, help="distance between variants to merge"
    )
    haplomerge_opt.add_argument(
        "-l", "--minlength", type=int, default=50, help="Minimum length of variants to consider"
    )
    haplomerge_opt.add_argument(
        "-c",
        "--callers",
        type=int,
        default=1,
        help="Minimum number of callers to support a variant",
    )
    haplomerge_opt.add_argument(
        "-i",
        "--ignore_type",
        help="Ignore the type of the structural variant",
        action="store_true",
        default=False,
    )
    haplomerge_opt.add_argument(
        "-s", "--strand", action="store_true", default=False, help="Take strand into account"
    )
    haplomerge_opt.add_argument(
        "-e",
        "--estimate_distance",
        action="store_true",
        default=False,
        help="Estimate distance between calls",
    )

    lengthplot = subparsers.add_parser(
        "lengthplot",
        help="create stacked bar plot of SV lengths split by type",
        parents=[parent_parser],
    )
    lengthplot_req = lengthplot.add_argument_group("required arguments")
    lengthplot_req.add_argument("vcf", help="vcf file to parse")
    lengthplot_opt = lengthplot.add_argument_group("optional arguments")
    lengthplot_opt.add_argument(
        "--plotout", help="output file to write figure to", default="SV-length.png"
    )
    lengthplot_opt.add_argument(
        "-c", "--counts", help="output file to write counts to", default="SV-length.txt"
    )
    lengthplot_opt.add_argument(
        "--all", help="Plot all variants and not just the first in the file", action="store_true"
    )
    minlength = subparsers.add_parser(
        "minlen", help="filter a SV vcf file by minimal variant length", parents=[parent_parser]
    )
    minlength_req = minlength.add_argument_group("required arguments")
    minlength_req.add_argument("vcf", help="vcf file to parse")

    minlength_opt = minlength.add_argument_group("optional arguments")
    minlength_opt.add_argument("-l", "--length", help="minimal SV length", type=int, default=50)
    minlength_opt.add_argument("-o", "--output", help="vcf file to write to", default=None)

    truncate_svlen = subparsers.add_parser(
        "svlentruncate",
        help="limit the SVLEN to a certain (positive) length",
        parents=[parent_parser],
    )
    truncate_svlen_req = truncate_svlen.add_argument_group("required arguments")
    truncate_svlen_req.add_argument("vcf", help="vcf file to parse")

    truncate_svlen_opt = truncate_svlen.add_argument_group("optional arguments")
    truncate_svlen_opt.add_argument(
        "-l",
        "--length",
        help="maximal SVLEN, replace SVLEN by this value if larger",
        type=int,
        default=1e5,
    )
    truncate_svlen_opt.add_argument("-o", "--output", help="vcf file to write to", default=None)

    fixvcf = subparsers.add_parser(
        "fixvcf", help="Some fixes to make compatible with e.g. vcfanno", parents=[parent_parser]
    )
    fixvcf_req = fixvcf.add_argument_group("required arguments")
    fixvcf_req.add_argument("vcf", help="vcf file to parse")
    fixvcf_req.add_argument("--fai", help="index of corresponding fasta file", required=True)
    fixvcf_opt = fixvcf.add_argument_group("optional arguments")
    fixvcf_opt.add_argument("-o", "--output", help="vcf file to write to", default="stdout")
    fixvcf_opt.add_argument(
        "--jasmine", help="Fix problems related to using jasmine", action="store_true"
    )

    purge2d = subparsers.add_parser(
        "purge2d", help="Remove accidental 2D reads from a bam file", parents=[parent_parser]
    )
    purge2d_req = purge2d.add_argument_group("required arguments")
    purge2d_req.add_argument("bam", help="bam file to filter")

    purge2d_opt = purge2d.add_argument_group("optional arguments")
    purge2d_opt.add_argument(
        "-o", "--output", help="sam/bam file to write filtered alignments to [stdout]", default="-"
    )

    carrierplot = subparsers.add_parser(
        "carrierplot", help="show number of carriers per variant", parents=[parent_parser]
    )
    carrierplot_req = carrierplot.add_argument_group("required arguments")
    carrierplot_req.add_argument("variants", help="VCF to plot from")

    carrierplot_opt = carrierplot.add_argument_group("optional arguments")
    carrierplot_opt.add_argument(
        "--plotout", help="output file to write figure to", default="SV-carriers.png"
    )

    varcount = subparsers.add_parser(
        "varcount", help="plot number of variants per sample in a vcf file", parents=[parent_parser]
    )
    varcount_req = varcount.add_argument_group("required arguments")
    varcount_req.add_argument("variants", help="VCF to plot from")

    varcount_opt = varcount.add_argument_group("optional arguments")
    varcount_opt.add_argument(
        "--plotout", help="output file to write figure to", default="SV-counts.png"
    )
    varcount_opt.add_argument(
        "--countsout", help="output file to write counts to", default="SV-counts.txt"
    )

    fixref = subparsers.add_parser(
        "fixref", help="Fix reference allele in vcf file", parents=[parent_parser]
    )
    fixref_req = fixref.add_argument_group("required arguments")
    fixref_req.add_argument("variants", help="VCF to fix")
    fixref_req.add_argument("--fasta", help="fasta file", required=True)

    args = parser.parse_args()
    validate_args(parser, args)
    return args


def validate_args(parser, args):
    if not args.command:
        sys.stderr.write("INPUT ERROR: sub-command required\n\n")
        parser.print_help()
        sys.exit()
    if args.command in ["upset", "venn"]:
        if args.names:
            if not len(args.variants) == len(args.names):
                sys.exit(
                    "INPUT ERROR: " "Need to have same number of values in --names as --variants!"
                )
    if args.command == "venn":
        if len(args.variants) > 3:
            sys.exit("INPUT ERROR: " "Venn diagrams are only created for 2 or 3 vcf files!")
    if args.command == "haplomerge":
        if not len(args.variants) in [2, 3]:
            sys.exit("INPUT ERROR: " "haplomerge can only be used on 2 or 3 vcf files!")
    if hasattr(args, "variants"):
        if isinstance(args.variants, list):
            for f in args.variants:
                if not path.isfile(f):
                    sys.exit(f"File not found: {f}")
        else:
            if not path.isfile(args.variants):
                sys.exit(f"File not found: {args.variants}")
    if hasattr(args, "truth"):
        if not path.isfile(args.truth):
            sys.exit("File not found: {}".format(args.truth))
    if hasattr(args, "test"):
        if not path.isfile(args.test):
            sys.exit("File not found: {}".format(args.test))


def get_survivor_version():
    import subprocess

    for line in subprocess.check_output(
        args="SURVIVOR", stderr=subprocess.STDOUT, universal_newlines=True
    ).split("\n"):
        if line.startswith("Version:"):
            return line.strip().split(" ")[1]
    else:
        return "version not found"



import pysam
import sys
from cigar import Cigar
import pandas as pd


def process(bamfile, output="-", write_candidates=False):
    twod = []
    bam = pysam.AlignmentFile(bamfile, "rb")
    mode = 'wb' if output.endswith('.bam') else 'w'
    filtered_alignments = pysam.AlignmentFile(output, mode, template=bam)
    for read in bam.fetch():
        if is_accidental_2d(read):
            twod.append(read)
        else:
            filtered_alignments.write(read)
    sys.stderr.write(f"Detected {len(twod)} potential artefacts "
                     f"out of {bam.mapped} alignments ({100*(len(twod))/bam.mapped}%)\n")
    if write_candidates:
        twod_bam = pysam.AlignmentFile("2D-candidates.bam", "wb", template=bam)
        for r in twod:
            twod_bam.write(r)
    keep_these_candidates = follow_up_2d_candidates(twod, distance=500)
    if keep_these_candidates:
        sys.stderr.write("WARNING: Some potential artefacts are close to another.\n")
        sys.stderr.write("WARNING: As this could be an SV, these reads are kept.\n")
        sys.stderr.write("WARNING: Surpyvor is producing an UNSORTED sam/bam.\n")
        for read in keep_these_candidates:
            filtered_alignments.write(read)


def get_sa_attributes(sa_tag):
    '''
    An entry in the SA tag consist of rname, POS, strand, CIGAR, mapQ, NM
    Returned are POS, strand and aligned length (based on cigar)
    '''
    sasplit = sa_tag.split(',')
    start = int(sasplit[1])
    end = start + Cigar(sasplit[3]).reference_length()
    strand = sasplit[2]
    return start, end, strand


def get_strand(read):
    return '-' if read.is_reverse else '+'


def reverse_overlaps(sa_start, sa_end, sa_strand, read):
    """
    Return True for reads overlapping their supplementary alignment on the reverse strand
    """
    return sa_strand != get_strand(read) \
        and max(sa_start, read.reference_start) <= min(sa_end, read.reference_end)


def is_accidental_2d(read):
    """
    Return True for supplementary alignments for which a single other (the primary) alignment exists
    for which the strand is reverse and the reference span is overlapping
    """
    if read.has_tag('SA') and read.is_supplementary:
        supps = [get_sa_attributes(s) for s in read.get_tag('SA').split(';') if s]
        if len(supps) > 1:
            return False
        start_sa, end_sa, strand_sa = supps[0]
        if reverse_overlaps(*supps[0], read):
            return True
        else:
            return False
    else:
        return False


def follow_up_2d_candidates(candidates, distance=500):
    df = pd.DataFrame.from_records(
        data=[(r.query_name, r.reference_name, r.reference_start, r.reference_end)
              for r in candidates],
        columns=["RNAME", "Chromosome", "Start", "End"])
    keep_these_reads = []
    for chrom in df["Chromosome"].unique():
        keep_these_reads.extend(
            follow_up_2d_candidates_per_chrom(
                df[df["Chromosome"] == chrom].sort_values(by=["Start", "End"]),
                distance=distance)
        )
    return [r for r in candidates if r.query_name in keep_these_reads]


def follow_up_2d_candidates_per_chrom(df, distance=500):
    """
    Per chromosome, calculate the minimum of the distance with the previous and next read
    Distances are calculated based on read start position.
    Return only those read identifiers that are <= <distance> from each other.
    """
    df["distance"] = pd.concat([df["Start"].diff().abs(),
                                df["Start"].diff(periods=-1).abs()],
                               axis="columns").min(axis='columns')
    return df.loc[df["distance"] < distance, "RNAME"].to_list()



from cyvcf2 import VCF, Writer
from pyfaidx import Fasta


def fixref(vcf, fasta):
    """
    Fix reference alleles in VCF file.

    Parameters
    ----------
    vcf : str
        Path to VCF file.
    fasta : str
        Path to FASTA file.
    """
    fas = Fasta(fasta)
    vcf = VCF(vcf)
    w = Writer("-", vcf)
    for v in vcf:
        v.REF = fas[v.CHROM][v.start : v.end].seq
        w.write_record(v)



import os
import sys
import tempfile
from cyvcf2 import VCF, Writer
import subprocess
import shlex
import pandas as pd


def is_variant(call):
    """Check if a variant position qualifies as a variant

    0,1,2,3==HOM_REF, HET, UNKNOWN, HOM_ALT"""
    if call == 1 or call == 3:
        return True
    else:
        return False


def normalize_vcf(vcff):
    """Normalize a vcf by changing DUP to INS"""
    import gzip

    handle, name = tempfile.mkstemp(suffix='.vcf')
    out = open(name, 'w')
    if vcff.endswith('.gz'):
        vcf = gzip.open(vcff, 'rt')
    else:
        vcf = open(vcff)
    for line in vcf:
        out.write(line.replace('DUP', 'INS'))
    os.close(handle)
    return name


def get_variant_identifiers(vcf, ignore_chroms, num_samples=2):
    """Get sets of variants for each sample in a merged vcf.

    Loop over the vcf file, adding a unique identifier to the
    respective list if the sample has a variant for that position
    return as set
    """
    positions = [[] for _ in range(num_samples)]
    for v in VCF(vcf):
        if v.CHROM not in ignore_chroms:
            for index, call in enumerate(v.gt_types):
                if is_variant(call):
                    positions[index].append(
                        "{}:{}-{}".format(v.CHROM, v.start, v.INFO.get('SVTYPE')))
    identifier_sets = [set(i) for i in positions]
    return identifier_sets


def gt_types_to_binary_comparison(calls):
    """From an array of calls, check if a variant position qualifies as a variant.

    0,1,2,3==HOM_REF, HET, UNKNOWN, HOM_ALT
    Return string of 1s and 0s to represent position"""
    binary_calls = []
    for call in calls:
        if call == 1 or call == 3:
            binary_calls.append(1)
        else:
            binary_calls.append(0)
    return ''.join([str(i) for i in binary_calls])


def make_sets(vcf, names):
    """From the merged SV file, return pd.Series of overlapping sets.

    Intended for making an upset plot"""
    from collections import defaultdict
    calls = defaultdict(int)
    for v in VCF(vcf):
        calls[gt_types_to_binary_comparison(v.gt_types)] += 1
    tf_array = [[True, False]] * len(list(calls.keys())[0])
    index = pd.MultiIndex.from_product(tf_array, names=names)
    values = [calls[''.join([str(int(j)) for j in i])] for i in index]
    return pd.Series(values, index=index)


def vcf_concat(vcffiles):
    _, concatenated = tempfile.mkstemp(suffix=".vcf")
    sample = [get_sample(f) for f in vcffiles][0]
    vcffiles = [reheader(f, sample=sample) for f in vcffiles]
    vcffiles = [compress_and_tabix(f) for f in vcffiles]
    c = subprocess.Popen(shlex.split("bcftools concat -a {}".format(' '.join(vcffiles))),
                         stdout=subprocess.PIPE)
    subprocess.call(shlex.split("bcftools sort -o {}".format(concatenated)), stdin=c.stdout)
    return concatenated


def get_sample(vcffile):
    vcf = VCF(vcffile)
    return vcf.samples[0]


def reheader(vcf, sample):
    _, output = tempfile.mkstemp(suffix=".vcf")
    handle, samplef = tempfile.mkstemp()
    open(samplef, 'w').write(sample)
    os.close(handle)
    subprocess.call(shlex.split("bcftools reheader -s {} {} -o {}".format(samplef, vcf, output)))
    return output


def compress_and_tabix(vcf):
    if vcf.endswith('.vcf'):
        handle, output = tempfile.mkstemp(suffix=".vcf.gz")
        subprocess.call(shlex.split("bgzip -c {}".format(vcf)), stdout=handle)
        subprocess.call(shlex.split("tabix -p vcf {}".format(output)))
        return output
    else:
        return vcf


def decompress(vcf):
    """
    Decompress output to temporary file if filename endswith .gz or .bgz
    """
    if vcf.endswith(('.gz', '.bgz')):
        handle, output = tempfile.mkstemp(suffix=".vcf")
        subprocess.call(shlex.split("bgzip -cd {}".format(vcf)), stdout=handle)
        return output
    else:
        return vcf


def test_dependencies():
    from shutil import which

    for dependency in ['bcftools', 'bgzip', 'tabix', 'SURVIVOR']:
        if not which(dependency):
            sys.exit("ERROR: Could not find required executable '{}'.\n"
                     "Make sure it is installed and in $PATH".format(dependency))


def vcf_sort(input, output):
    if output in ["stdout", "-"]:
        subprocess.call(shlex.split('bcftools sort {}'.format(input)))
    else:
        subprocess.call(shlex.split('bcftools sort {} -o {}'.format(input, output)))


def confusion_matrix(vcff, names):
    """
    First level of the dict is the "first" call, second level is the "second" sample
    0: hom_ref
    1: heterozygous
    2: unknown/nocall
    3: hom_alt
    """
    zygosities = {0: {0: 0, 1: 0, 2: 0, 3: 0},
                  1: {0: 0, 1: 0, 2: 0, 3: 0},
                  2: {0: 0, 1: 0, 2: 0, 3: 0},
                  3: {0: 0, 1: 0, 2: 0, 3: 0},
                  }
    for v in VCF(vcff):
        zygosities[v.gt_types[0]][v.gt_types[1]] += 1
    zygs = [2, 0, 1, 3]
    df = pd.DataFrame(index=zygs, columns=zygs)
    for tr in zygs:
        for te in zygs:
            df.loc[tr, te] = zygosities[tr][te]
    df.columns = ['nocall', 'hom_ref', 'het', 'hom_alt']
    df.columns.name = names[1]
    df.index = ['nocall', 'hom_ref', 'het', 'hom_alt']
    df.index.name = names[0]
    print(df)


def get_svlengths(vcf, all=False):
    from collections import defaultdict
    len_dict = defaultdict(list)
    vcf = VCF(vcf)
    if not all and len(vcf.samples) > 1:
        sys.stderr.write(f"\n\nWarning: only using first sample: {vcf.samples[0]}.\n")
        sys.stderr.write("Use --all to plot all variants.\n")
    for v in vcf:
        if all or is_variant(v.gt_types[0]) and not v.INFO.get('SVTYPE') == 'TRA':
            try:
                if get_svlen(v) >= 50:
                    len_dict[get_svtype(v)].append(get_svlen(v))
            except TypeError:
                if v.INFO.get('SVTYPE') == 'INV':
                    if (v.end - v.start) >= 50:
                        len_dict[get_svtype(v)].append(v.end - v.start)
                elif v.INFO.get('SVTYPE') == 'BND':
                    try:
                        len_dict['BND'].append(abs(v.INFO.get('MATEDIST')))
                    except TypeError:
                        len_dict['BND'].append(0)
                else:
                    sys.stderr.write("Exception when parsing variant:\n{}\n\n".format(v))
                    len_dict["parse_error"].append(0)
    return len_dict


def get_svtype(v):
    if v.INFO.get('SVTYPE') == "INVDUP":
        return "INV"
    else:
        return v.INFO.get('SVTYPE').split(':')[0].split('/')[0]


def get_svlen(v):
    return abs(v.INFO.get('SVLEN'))


def filter_vcf(vcf, output, minlength=0, truncate_svlen=float("inf"), suffix=""):
    vcf_in = VCF(vcf)
    if not output:
        output = vcf.replace(".vcf", "_{}.vcf".format(suffix))
    vcf_in.add_info_to_header({'ID': 'TRUNCATED',
                               'Description': "SVLEN truncated",
                               'Type': 'Flag', 'Number': '0'})
    vcf_out = Writer(output, vcf_in)
    records_truncated = 0
    records_filtered = 0
    for v in vcf_in:
        svlen = get_svlen(v)
        if svlen >= minlength:
            if svlen > truncate_svlen:
                v.INFO['SVLEN'] = 1
                v.INFO['END'] = v.start + 1
                v.INFO['TRUNCATED'] = True
                records_truncated += 1
            vcf_out.write_record(v)
        else:
            records_filtered += 1
    if records_truncated != 0:
        sys.stderr.write("Truncated {} records where SVLEN > {}\n".format(
            records_truncated, int(truncate_svlen)))
    if records_filtered != 0:
        sys.stderr.write("Filtered {} records where SVLEN < {}\n".format(
            records_filtered, int(minlength)))


def fix_vcf(vcf, output, fai, jasmine=False):
    chromsizes = {line.split()[0]: int(line.split()[1]) for line in open(fai)}

    vcf_in = VCF(vcf)
    if not output:
        output = vcf.replace(".vcf", "_{}.vcf".format("fixed"))
    vcf_in.add_info_to_header({'ID': 'TRUNCATED',
                               'Description': "SVLEN truncated",
                               'Type': 'Flag', 'Number': '0'})
    vcf_in.add_info_to_header({'ID': 'STRANDS2',
                               'Description': "alt reads first +,alt reads first -,alt reads second +,alt reads second -.",
                               'Type': 'Integer', 'Number': '4'})
    vcf_in.add_info_to_header({'ID': 'Strandbias_pval',
                               'Description': "P-value for fisher exact test for strand bias.",
                               'Type': 'Float', 'Number': 'A'})
    vcf_in.add_filter_to_header({'ID': 'STRANDBIAS',
                                 'Description': "Strand is biased if Strandbias_pval< 0.01."})
    if jasmine:
        vcf_in.add_info_to_header({'ID': 'STRANDS',
                                   'Description': "foo",
                                   'Type': 'String', 'Number': '1'})
        vcf_in.add_info_to_header({'ID': 'AF',
                                   'Description': "foo",
                                   'Type': 'Float', 'Number': '1'})
    handle, interm_output = tempfile.mkstemp(suffix=".vcf")
    vcf_out = Writer(interm_output, vcf_in)
    records_fixed = 0
    records_truncated = 0
    mito_variants = 0
    interchromosomal_bnds = 0
    for v in vcf_in:
        if v.CHROM == 'chrM':
            mito_variants += 1
            continue
        if v.start == -1:
            v.set_pos(0)
            records_fixed += 1
        try:
            if (v.INFO.get('SVTYPE') == 'BND') and (v.CHROM != v.INFO.get('CHR2')):
                del v.INFO['END']
                interchromosomal_bnds += 1
        except KeyError:
            pass
        try:
            if chromsizes[v.INFO.get('CHR2')] < v.INFO.get('END'):
                v.INFO['SVLEN'] = 1
                v.INFO['END'] = v.start + 1
                v.INFO['TRUNCATED'] = True
                records_truncated += 1
        except KeyError:
            pass
        if v.INFO.get('SVLEN') == 999999999:
            v.INFO['SVLEN'] = 1
            v.INFO['TRUNCATED'] = True
        vcf_out.write_record(v)
    vcf_out.close()
    vcf_sort(interm_output, output)
    if mito_variants != 0:
        sys.stderr.write(f"Removed {mito_variants} records on chrM.\n")
    if records_fixed != 0:
        sys.stderr.write(f"Fixed {records_fixed} records.\n")
    if records_truncated != 0:
        sys.stderr.write(f"Truncated {records_truncated} records where END > chromosome size\n")
    if interchromosomal_bnds != 0:
        sys.stderr.write(f"Dropped END for {interchromosomal_bnds} interchromosomal BNDs\n")



import tempfile
from cyvcf2 import VCF
from surpyvor import utils
import sys


def merge_split_called_haplotypes(merged, output, name=None):
    _, name = tempfile.mkstemp(suffix='.vcf')
    vcf = VCF(merged)
    if len(vcf.samples) == 3:
        get_genotype = get_genotype_from_three
    elif len(vcf.samples) == 2:
        get_genotype = get_genotype_from_two
    else:
        sys.exit("ERROR: Unexpected number of samples in haplomerge intermediate VCF!")
    with open(name, 'a') as tmpoutput:
        tmpoutput.write("{}\n".format('\n'.join(make_header(vcf, name=name))))
        allele_dict = {0: 'HOM_REF', 1: 'HET', 2: 'UNKNOWN', 3: 'HOM_ALT'}
        for v in vcf:
            info = {'SVLEN': v.INFO.get('SVLEN'),
                    'END': v.end,
                    'SVTYPE': v.INFO.get('SVTYPE'),
                    'HAPSUPPORT': '-'.join([allele_dict[gt] for gt in v.gt_types])}
            print("{chrom}\t{pos}\t{idf}\t{ref}\t{alt}\t{q}\t{filt}\t{info}\t{form}\t{sam}"
                  .format(
                      chrom=v.CHROM,
                      pos=v.start,
                      idf=v.ID,
                      ref=v.REF,
                      alt=','.join(v.ALT),
                      q=v.QUAL or '.',
                      filt='.',
                      info=';'.join(['{}={}'.format(k, v) for k, v in info.items()]),
                      form='GT',
                      sam=get_genotype(v.gt_types)),
                  file=tmpoutput)
    utils.vcf_sort(name, output)


def get_genotype_from_two(alleles):
    '''Combine calls on two haplotypes in one genotype'''
    hap1, hap2 = [int(utils.is_variant(c)) for c in alleles]
    return '{}|{}'.format(hap1, hap2)


def get_genotype_from_three(alleles):
    '''Combine calls on two haplotypes and unphased set in one genotype'''
    calls = [int(utils.is_variant(c)) for c in alleles]
    if sum(calls) >= 2 or calls == [0, 0, 1]:
        return '1|1'
    elif sum(calls) == 1:
        return '{}|{}'.format(calls[0], calls[1])
    else:  # I don't think should happen
        return '0|0'


def make_header(vcf, name=None):
    header = ['##fileformat=VCFv4.1', '##source=surpyvor haplomerge']
    for line in vcf.header_iter():
        if line["HeaderType"] == 'CONTIG':
            header.append('##contig=<ID={}>'.format(line['ID']))
    if name is None:
        name = "SAMPLE"
    header.extend([
        '##ALT=<ID=DEL,Description="Deletion">',
        '##ALT=<ID=DUP,Description="Duplication">',
        '##ALT=<ID=INV,Description="Inversion">',
        '##ALT=<ID=BND,Description="Translocation">',
        '##ALT=<ID=INS,Description="Insertion">',
        '##INFO=<ID=END,Number=1,Type=Integer,Description="End of the structural variant">',
        '##INFO=<ID=SVLEN,Number=1,Type=Float,Description="Length of the SV">',
        '##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of the SV.">',
        '##INFO=<ID=HAPSUPPORT,Number=1,Type=String,Description="Support per haplotype.">',
        '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">',
        '#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t{}'.format(name)])
    return header



import matplotlib.pyplot as plt


def bar_chart(vcf, outname="stacked_bar.png"):
    """
    Make a stacked bar chart for length of the SV split by validation status
    This ignores zygosity.
    """
    from cyvcf2 import VCF
    from surpyvor import utils
    import numpy as np

    len_dict = {"True": [], "False": [], "Missed": []}
    for v in VCF(vcf):
        if not v.INFO.get("SVTYPE") == "TRA" and abs(v.INFO.get("SVLEN")) >= 50:
            calls = [utils.is_variant(call) for call in v.gt_types]
            if calls == [True, True]:
                len_dict["True"].append(v.INFO.get("SVLEN"))
            elif calls == [False, True]:
                len_dict["False"].append(v.INFO.get("SVLEN"))
            elif calls == [True, False]:
                len_dict["Missed"].append(v.INFO.get("SVLEN"))
    plt.subplot(2, 1, 1)
    plt.hist(
        x=np.array(list(len_dict.values())),
        bins=[i for i in range(0, 2000, 10)],
        stacked=True,
        histtype="bar",
        label=list(len_dict.keys()),
    )
    plt.xlabel("Length of structural variant")
    plt.ylabel("Number of variants")
    plt.legend(frameon=False, fontsize="small")
    plt.subplot(2, 1, 2)
    plt.hist(
        x=np.array(list(len_dict.values())),
        bins=[i for i in range(0, 20000, 100)],
        stacked=True,
        histtype="bar",
        label=list(len_dict.keys()),
        log=True,
    )
    plt.xlabel("Length of structural variant")
    plt.ylabel("Number of variants")
    plt.legend(frameon=False, fontsize="small")
    plt.tight_layout()
    plt.savefig(outname)
    plt.close()


def num_variants_per_sample(vcf, outname="num_variants_per_sample.png", counts_out="counts.txt"):
    """
    Make a scatter plot of the number of variants per sample
    """
    from cyvcf2 import VCF
    import numpy as np

    vcf = VCF(vcf)
    calls = np.zeros(len(vcf.samples))
    for v in vcf:
        calls += np.array([int(g in [1, 3]) for g in v.gt_types])
    ids = vcf.samples
    # sort the counts and ids by counts
    counts, ids = zip(*sorted(zip(calls.tolist(), ids), reverse=True))
    with open(counts_out, "w") as out:
        for i, c in zip(ids, counts):
            out.write(f"{i}\t{c}\n")
    plt.scatter(x=ids, y=counts, s=5)
    plt.xlabel("Samples")
    if len(ids) > 20:
        plt.xticks([])
    plt.ylabel("Number of variants")
    plt.tight_layout()
    plt.savefig(outname)
    plt.close()


def upset_plot(upsets, outname="UpSetPlot.png"):
    from upsetplot import plot as upsetplot

    upsetplot(upsets, sort_by="cardinality")
    plt.savefig(outname)


def venn_diagram(sets, labels, num_samples=2, outname="venn.png"):
    if num_samples == 2:
        from matplotlib_venn import venn2 as venn
    else:
        from matplotlib_venn import venn3 as venn
    venn(sets, set_labels=labels)
    plt.savefig(outname)
    plt.close()


def length_plot(dict_of_lengths, output):
    """Makes two stacked bar charts
    Plotting two bar charts of number of SVs by length split by SV type
    Use a consistent colouring scheme for those in "standard_order" to
    make comparison reasonable

    First bar chart is up to 2kb with bins of 10bp
    Second bar chart is up to 20kb, with bins of 100bp
     and uses log scaling on the y-axis
    """
    standard_order = ["DEL", "INS", "INV", "DUP"]
    spec_order = sorted([i for i in dict_of_lengths.keys() if i not in standard_order])
    sorter = standard_order + spec_order
    names, lengths = zip(
        *sorted(
            [(svtype, lengths) for svtype, lengths in dict_of_lengths.items()],
            key=lambda x: sorter.index(x[0]),
        )
    )
    plt.subplot(2, 1, 1)
    plt.hist(
        x=lengths, bins=[i for i in range(50, 2000, 10)], stacked=True, histtype="bar", label=names
    )
    plt.xlabel("Length of structural variant")
    plt.ylabel("Number of variants")
    plt.legend(frameon=False, fontsize="small")

    plt.subplot(2, 1, 2)
    plt.hist(
        x=lengths,
        bins=[i for i in range(0, 20000, 100)],
        stacked=True,
        histtype="bar",
        label=names,
        log=True,
    )
    plt.xlabel("Length of structural variant")
    plt.ylabel("Number of variants")
    plt.legend(frameon=False, fontsize="small")
    plt.tight_layout()
    plt.savefig(output)


def carrierplot(args):
    from cyvcf2 import VCF
    from surpyvor.utils import is_variant
    import numpy as np

    vcf = VCF(args.variants)
    counts = np.array([sum([is_variant(call) for call in v.gt_types]) for v in vcf])
    plt.hist(x=counts, bins=[i for i in range(1, len(vcf.samples))], histtype="bar")
    plt.xlabel("Number of carriers")
    plt.ylabel("Number of variants")
    plt.tight_layout()
    plt.savefig(args.plotout)



# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
exec(open("surpyvor/version.py").read())

setup(
    name="surpyvor",
    version=__version__,
    description="Manipulate vcf files of structural variants using SURVIVOR",
    long_description=open(path.join(here, "README.md")).read(),
    long_description_content_type="text/markdown",
    url="https://github.com/wdecoster/surpyvor",
    author="Wouter De Coster",
    author_email="decosterwouter@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],
    keywords="nanopore",
    packages=find_packages(),
    python_requires=">=3",
    install_requires=[
        "matplotlib",
        "matplotlib-venn",
        "cyvcf2==0.30.18",
        "numpy",
        "UpSetPlot",
        "pysam",
        "cigar",
        "pyfaidx",
    ],
    package_data={"surpyvor": []},
    package_dir={"surpyvor": "surpyvor"},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "surpyvor=surpyvor.surpyvor:main",
        ],
    },
)
